import os
import sys

import click

from .filterspec import FilterSpec
from .seeqrypt.nacl_backend import (
    load_private_key,
)
from .storage.sqlite_storage import SqliteStorage


def import_secrets(sender, file, value, serializer):
    storage = SqliteStorage()
    receiver = storage.fetch_admin()
    sender = storage.fetch_users(username=sender)[0]

    s = serializer(
        sender=sender,
        receiver=receiver,
        receiver_private_key=load_private_key('private.key'),
    )
    secrets = s.load(file or value)
    for secret in secrets:
        storage.add_secret(secret)


def export_secrets(ctx, *, to: str, fspec: FilterSpec, serializer,
                   out=None, windows=False, linux=False):
    """seeqret export <user>

       Exports secrets from a SQLite database, preparing them for secure
       transfer to the specified receiver. The function retrieves secrets,
       decrypts them with a symmetric key, then encrypts the values with
       asymmetric encryption for the receiver. Finally, it generates a
       signature for the exported secrets for authenticity and integrity.

       Parameters:
           to (str): The recipient's username or the string 'self'
                     indicating the administrator. Secrets will be encrypted
                     using the public key of the specified user or
                     administrator.

       Returns:
           dict: A dictionary structured to contain metadata of the transaction
                 including the sender, receiver, and encrypted secrets ready
                 for transfer, along with a signature.

       Raises:
           Various exceptions may arise during database connection issues,
           encryption/decryption processes, or key-loading operations,
           depending on implementation details not shown here.
    """
    storage = SqliteStorage()

    admin = storage.fetch_admin()
    if to == 'self':
        receiver = admin
    else:
        receiver = storage.fetch_users(username=to)[0]

    s = serializer(
        sender=admin,
        receiver=receiver,
        sender_private_key=load_private_key('private.key'),
    )

    secrets = storage.fetch_secrets(**fspec.to_filterdict())
    system = sys.platform  # default to current system
    if windows:
        system = 'win32'
    if linux:
        system = 'linux'
    res = s.dumps(secrets, system)

    if out:
        click.echo(f"Writing secrets to: {out} ({os.path.join(ctx.obj['curdir'], out)})")
        with open(os.path.join(ctx.obj['curdir'], out), 'w') as f:
            f.write(res)
    else:
        click.echo(res)
    return res



from click.testing import CliRunner

from seeqret.db_utils import debug_fetch_users, debug_secrets
from seeqret.main import cli, user, users, init, list, get
from seeqret.cli_group_add import key, text as add_text
from seeqret.storage.get_secret import get_secret
from tests.clirunner_utils import print_result


def test_add_key():
    runner = CliRunner(env=dict(TESTING="TRUE"))
    with runner.isolated_filesystem():
        result = runner.invoke(init, [
            '.',
            '--user=test',
            '--email=test@example.com',
        ])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0

        assert len(debug_secrets()) == 0
        result = runner.invoke(list)
        assert result.exit_code == 0
        assert len(debug_fetch_users()) == 1
        
        result = runner.invoke(key, [
            'FOO', 'BAR',
            '--app=myapp',
            '--env=dev'
        ])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0
        assert len(debug_secrets()) == 1

        result = runner.invoke(get, ['myapp:dev:FOO'])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0
        assert result.output == 'BAR\n'



def test_add_text():
    runner = CliRunner(env=dict(TESTING="TRUE"))
    with runner.isolated_filesystem():
        result = runner.invoke(init, [
            '.',
            '--user=test',
            '--email=test@example.com',
        ])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0

        text_input = '1234567890\n1234567890\n1234567890\n'

        result = runner.invoke(add_text, [
            'FOO',
            '--app=myapp',
            '--env=dev'
        ], input=text_input + '\x04')  # add EOF
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0

        result = runner.invoke(get, ['myapp:dev:FOO'])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0
        assert result.output == text_input + '\n'


# def test_add_file():
#     runner = CliRunner(env=dict(TESTING="TRUE"))
#     with runner.isolated_filesystem():
#         result = runner.invoke(init, [
#             '.',
#             '--user=test',
#             '--email=test@example.com',
#         ])
#         if result.exit_code != 0: print_result(result)
#         assert result.exit_code == 0

#         with open('test.txt', 'w') as f:
#             f.write('a = "1234567890"\nb = "1234567890"\nc = "1234567890"')

#         result = runner.invoke(add_env, [
#             'test.txt',
#             '--app=myapp',
#             '--env=dev'
#         ])
#         if result.exit_code != 0: print_result(result)
#         assert result.exit_code == 0

#         print("DEBUG_SECRETS", debug_secrets())
#         result = runner.invoke(get, ['myapp:dev:a'])
#         if result.exit_code != 0: print_result(result)
#         assert result.exit_code == 0
#         assert result.output == '1234567890'



import collections
import json
import sys

import pytest
import datetime
import decimal

from seeqret.models import jason


def roundtrip(v):
    "Convenience function to thest the roundtrip (dump/eval)."
    return jason.json_eval(jason.dumps(v)) == v


def test_jason_eval():
    "Test the jason_eval function using the roundtrip convenience function."
    assert roundtrip([])
    assert roundtrip(['hello world'])
    assert roundtrip(['hello world'.split()])
    assert roundtrip({})


def test_dumps():
    "Test the dumps function."
    dt = datetime.datetime(2012, 4, 2, 6, 12)
    assert jason.dumps(dt, indent=None) == '"2012-04-02T06:12:00"'
    assert jason.dumps(decimal.Decimal('3.14159263')) == repr(float('3.14159263'))
    assert jason.dumps({}.keys()) == '[]'
    assert jason.dumps({}.values()) == '[]'
    assert jason.dumps(range(0)) == '[]'
    assert jason.dumps(u'hello'.encode('u8')) == '"hello"'


def test_generator():
    assert jason.dumps((i for i in range(0))) == '[]'


def test_dictviews():
    a42 = dict(a=42)
    assert jason.json_eval(jason.dumps(a42.keys())) == ['a']
    assert jason.json_eval(jason.dumps(a42.values())) == [42]


def test_class_dumps():
    """Test the dump of the jason value of a class by using
       the __jason__ method.
    """

    class C:
        def __json__(self):
            return 42

    assert jason.dumps(C()) == '42'

    class D(object):
        def __init__(self):
            self.a = 42

    assert jason.json_eval(jason.dumps(D())) == {"a": 42}


def test_set_dumps():
    assert jason.json_eval(jason.dumps(1)) == 1
    assert jason.json_eval(jason.dumps(set())) == []
    assert jason.json_eval(jason.dumps({1, 2})) == [1, 2]
    assert jason.dumps(datetime.date(2019, 3, 15)) == '"2019-03-15"'

    class Foo(object):
        __slots__ = ['a', 'b']

    with pytest.raises(TypeError):
        jason.dumps(Foo())  # not JSON serializable


def test_mapping():
    class Bar(collections.abc.Mapping):
        __slots__ = ['a']

        def __getitem__(self, item):  # pragma: nocover
            pass

        def __iter__(self):
            return iter([])

        def __len__(self):  # pragma: nocover
            return 0

    assert isinstance(Bar(), collections.abc.Mapping)
    assert jason.dumps(Bar()) == "{}"


def test_loads():
    val = '{"k":"1970-05-02T06:10:00"}'
    jval = jason.loads(val)
    assert jval['k'] == datetime.datetime(1970, 5, 2, 6, 10)

    val = '{"k":"1970-05-02"}'
    jval = jason.loads(val)
    assert jval['k'] == datetime.date(1970, 5, 2)

    val = '{"k":"1970-05-02"}'
    jval = jason.loads(val)
    assert isinstance(jval['k'], datetime.date)

    val = '{"k":"@date1970-05-02"}'
    jval = jason.loads(val)
    assert not isinstance(jval['k'], datetime.date)

    val = '{"k":42}'
    jval = jason.loads(val)
    assert jval['k'] == 42

    val = b'{"k":42}'
    jval = jason.loads(val)
    assert jval['k'] == 42

    import json
    val = b'{"k":42}'
    jval = jason.loads(val, cls=json.JSONDecoder)
    assert jval['k'] == 42


def test_jsonname():
    assert jason.jsonname("hello.world") == "hello_world"



"""
Test that all modules are importable.
"""

import seeqret
import seeqret.fileutils
import seeqret.main
import seeqret.run_utils
import seeqret.seeqret_add
import seeqret.seeqret_init
import seeqret.seeqret_transfer
import seeqret.seeqrypt
import seeqret.seeqrypt.aes_fernet
import seeqret.seeqrypt.nacl_backend
import seeqret.seeqrypt.utils


def test_import_seeqret():
    """Test that all modules are importable.
    """
    
    assert seeqret
    assert seeqret.fileutils
    assert seeqret.main
    assert seeqret.run_utils
    assert seeqret.seeqret_add
    assert seeqret.seeqret_init
    assert seeqret.seeqret_transfer
    assert seeqret.seeqrypt
    assert seeqret.seeqrypt.aes_fernet
    assert seeqret.seeqrypt.nacl_backend
    assert seeqret.seeqrypt.utils



import re
import sys

from click.testing import CliRunner
from seeqret.cli_group_add import key
from seeqret.main import value, init, get
from tests.clirunner_utils import print_result
from seeqret.run_utils import current_user

def test_edit_value():
    runner = CliRunner(env=dict(TESTING="TRUE"))
    with runner.isolated_filesystem():
        result = runner.invoke(init, [
            '.',
            '--user=' + current_user(),
            '--email=test@example.com',
        ])
        if result.exit_code != 0:  print_result(result)
        assert result.exit_code == 0

        result = runner.invoke(key, [
            'FOO', 'BAR',
            '--app=myapp',
            '--env=dev'
        ])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0
    
        result = runner.invoke(value, ['myapp:dev:FOO', 'BAZ'])
        if result.exit_code != 0:  print_result(result)
        assert result.exit_code == 0

        result = runner.invoke(get, ['myapp:dev:FOO'])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0
        assert result.output == 'BAZ\n'

        result = runner.invoke(key, [
            'FOO', 'QUUX',
            '--app=myapp',
            '--env=prod'
        ])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0

        result = runner.invoke(get, ['myapp:prod:FOO'])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0
        assert result.output == 'QUUX\n'

        result = runner.invoke(value, ['FOO', 'ZAAP', '--all'])
        if result.exit_code != 0: print_result(result)
        assert result.exit_code == 0



import click

# from seeqret.serializers.serializer import SERIALIZERS
from .filterspec import FilterSpec
from .models import Secret
from .storage.sqlite_storage import SqliteStorage
from .run_utils import seeqret_dir


@click.command()
@click.pass_context
@click.argument('name')
@click.option('--app', default='*', show_default=True,
              help='The app to add the secret to')
@click.option('--env', default='*', show_default=True,
              help='The env(ironment) to add the secret to (e.g. dev/prod)')
def text(ctx, name: str, app: str | None = None, env: str | None = None):  # noqa: F811
    """Add a new multi-line secret with key NAME, eg recovery codes.
       You will need to enter the secret value interactively.
    """
    with seeqret_dir():
        storage = SqliteStorage()
        fspec = FilterSpec(f'{app}:{env}:{name}')
        secrets = storage.fetch_secrets(**fspec.to_filterdict())
        if secrets:
            ctx.fail(click.style(
                f'Secret {", ".join(s.key for s in secrets)} already exists!',
                fg='red'
            ))
        click.secho(f"Enter the secret value for {name} in {app}:{env}. ", fg='blue')
        click.secho("Enter EOF (Ctrl-D) when done.", fg='blue')
        value = ''
        while True:
            ch = click.getchar(echo=True)
            if ch == '\x04':
                break
            if ch == '\r':
                ch = ''
            if ch == '\n':
                print()
            # sys.stdout.write(ch)
            value += ch

        secret = Secret(
            app=app or '*',
            env=env or '*',
            key=name,
            plaintext_value=value,
            type='str'
        )
        storage.add_secret(secret)
        click.secho(f"added {name} to {app}:{env}", fg='green')
        # as_table('app,env,key,value,type', [secret])


@click.command()
@click.pass_context
@click.argument('name')
@click.argument('value')
@click.option('--app', default='*', show_default=True,
              help='The app to add the secret to')
@click.option('--env', default='*', show_default=True,
              help='The env(ironment) to add the secret to (e.g. dev/prod)')
def key(ctx, name: str, value: str, app: str | None = None, env: str | None = None):  # noqa: F811
    """Add a new NAME -> VALUE mapping.

       You can (should) specify the app and environment properties when adding
       a new mapping.
    """
    if (':' in name
            or (isinstance(app, str) and ':' in app)
            or (isinstance(env, str) and ':' in env)):

        ctx.fail(click.style(
            'Colon `:` is not valid in key, app, or env', fg='red'
        ))

    click.secho(
        f'Adding a new key: {name}, value: {value}, app: {app}, env: {env}',
        fg='blue'
    )

    with seeqret_dir():
        storage = SqliteStorage()
        fspec = FilterSpec(f'{app}:{env}:{name}')
        secrets = storage.fetch_secrets(**fspec.to_filterdict())
        if secrets:
            ctx.fail(click.style(
                f'Secret {", ".join(s.key for s in secrets)} already exists!',
                fg='red'
            ))
        secret = Secret(
            app=app or '*',
            env=env or '*',
            key=name,
            plaintext_value=value,
            type='str'
        )
        storage.add_secret(secret)

        # verify that it worked
        secrets = storage.fetch_secrets(app=app, env=env, key=name)
        if secrets:
            click.secho(
                f'..successfully added: {app}:{env}[{key}]', fg='green'
            )
        else:
            ctx.fail(click.style(
                f'Error: {app}:{env}[{key}] not written to database', fg='red'
            ))


# @click.command()
# @click.pass_context
# @click.option('--app', default='*', show_default=True,
#               help='The app to add the secret to')
# @click.option('--env', default='*', show_default=True,
#               help='The env(ironment) to add the secret to (e.g. dev/prod)')
# @click.argument('file', type=click.File('r'))
# def env(ctx, app, env, file):
#     """Add a new FILE to the vault (.env file format).
#     """
#     serializer_cls = SERIALIZERS.get('env')
#     if not serializer_cls:
#         ctx.fail(click.style(
#             'No serializer found for .env files', fg='red'
#         ))
#     assert serializer_cls is not None
#     serializer = serializer_cls()
#     with seeqret_dir():
#         storage = SqliteStorage()
#         secrets = serializer.load(file.read(), app=app, env=env)
#         click.echo(f"SECRETS {secrets}")
#         for secret in secrets:
#             click.echo(f"ADDING:SECRET {secret}")
#             storage.add_secret(secret)



import re
import sys

from click.testing import CliRunner
from seeqret.main import cli, init, info
from seeqret.migrations.utils import current_version
from tests.clirunner_utils import print_result


def test_info():
    runner = CliRunner(env=dict(TESTING="TRUE"))
    with runner.isolated_filesystem():
        result = runner.invoke(info)
        if result.exit_code != 0:  print_result(result)
        assert result.exit_code == 0

        result = runner.invoke(info, ['--dump'])
        if result.exit_code != 0:  print_result(result)
        assert result.exit_code == 0



# import sqlite3

# from seeqret.seeqrypt.aes_fernet import decrypt_string
# from seeqret.seeqrypt.utils import load_symetric_key
# from seeqret.run_utils import seeqret_dir
from seeqret.storage.get_secret import get_secret

__version__ = '0.1.9'


def get(key, app='*', env='*'):
    """Get a value from the configuration.

    Args:
        key: The key to get.
        app: The application to get the key from.
        env: The environment to get the key from.

    Returns:
        The value of the key.
    """
    return get_secret(key, app, env)
    # with seeqret_dir():
    #     cipher = load_symetric_key('seeqret.key')
    #     # open the database in read-only mode
    #     cn = sqlite3.connect('file:seeqrets.db?mode=ro', uri=True)
    #     cursor = cn.cursor()
    #     cursor.execute(
    #         "select value from secrets where app = ? and env = ? and key = ?",
    #         (app, env, key)
    #     )
    #     result = cursor.fetchone()
    #     cursor.close()
    #     cn.close()
    #     return decrypt_string(cipher, result[0]) if result else None



import re
import sys

from click.testing import CliRunner
from seeqret.main import cli, init, validate_current_user
from seeqret.migrations.utils import current_version
from tests.clirunner_utils import print_result
from seeqret.run_utils import current_user


def test_init():
    runner = CliRunner(env=dict(TESTING="TRUE"))
    with runner.isolated_filesystem():
        result = runner.invoke(init, [
            '.',
            '--user=' + current_user(),
            '--email=test@example.com',
        ])
        if result.exit_code != 0:  print_result(result)
        assert result.exit_code == 0
        if sys.platform == 'win32':
            assert 'vault_dir permissions are ok' in result.output
            assert 'vault is encrypted' in result.output
        assert 'seeqret.key created' in result.output

        import sqlite3
        cn = sqlite3.connect('seeqret/seeqrets.db')
        assert current_version(cn) >= 2

        assert validate_current_user()


def test_init_no_dir():
    runner = CliRunner(env=dict(TESTING="TRUE"))

    result = runner.invoke(init, [
        '/this/path/does/not/exist',
        '--user=test',
        '--email=test@example.com',
    ])

    assert re.search(r'Error: The parent of the vault: .*? must exist.', result.output)



import os
import sys
from contextlib import contextmanager

import click


@contextmanager
def cd(path):
    """Change the current working directory.
    """
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


def get_seeqret_dir():
    if sys.platform == 'win32' or 'CI' in os.environ or 'SEEQRET' in os.environ:
        return os.environ['SEEQRET']
    else:
        return '/srv/.seeqret'


def is_initialized():
    if 'SEEQRET' not in os.environ:
        return False
    sdir = get_seeqret_dir()
    if not os.path.exists(sdir):
        return False
    if not os.path.exists(os.path.join(sdir, 'seeqrets.db')):
        return False
    return True


@contextmanager
def seeqret_dir():
    old_dir = os.getcwd()
    os.chdir(get_seeqret_dir())
    try:
        yield
    finally:
        os.chdir(old_dir)


def run(cmd, echo=True, workdir='.'):
    if echo:
        click.secho(f"    > {cmd}", fg='blue')
    with cd(workdir):
        res = os.popen(cmd).read()
    if echo:
        for line in res.split('\n'):
            click.secho(f"      {line}", fg='green')
    return res


def current_user():
    if sys.platform == 'win32':
        import ctypes
        GetUserNameExW = ctypes.windll.secur32.GetUserNameExW
        # https://learn.microsoft.com/en-us/windows/win32/api/secext/ne-secext-extended_name_format # noqa
        sam_compatible = 2  # MACHINE|domain\username (e.g. MYPC\thebjorn)

        size = ctypes.pointer(ctypes.c_ulong(0))
        GetUserNameExW(sam_compatible, None, size)

        name_buffer = ctypes.create_unicode_buffer(size.contents.value)
        GetUserNameExW(sam_compatible, name_buffer, size)
        domain_name = name_buffer.value
        return domain_name.split('\\')[1]
    else:
        import pwd
        # username (e.g. thebjorn)
        print("CURRENT USER", pwd.getpwuid(os.geteuid()).pw_name)
        return pwd.getpwuid(os.geteuid()).pw_name


if __name__ == '__main__':
    print(current_user())
